// import { Debug, UndefinedError } from './base';
import { Debug } from './base';
import { DefinedNumber } from './util/definedvalue';
import { Message, MessageCode } from "./util/message";

/**
 * domaine de définition du paramètre
 */
export enum ParamDomainValue {
    /**
     * >0, =0, <0 (-inf -> +inf)
     */
    ANY,

    /**
     * >=0
     */
    POS_NULL,


    /**
     * > 0
     */
    POS,

    /**
     * <>0
     */
    NOT_NULL,

    /**
     * intervalle
     */
    INTERVAL
}

export class ParamDomain {
    private _domain: ParamDomainValue;

    private _minValue: number;

    private _maxValue: number;

    constructor(d: ParamDomainValue, min: number = undefined, max: number = undefined) {
        this.checkValue(d, min, max);
        this._domain = d;
        this._minValue = min;
        this._maxValue = max;
    }

    private checkValue(val: number, min: number, max: number) {
        switch (val) {
            case ParamDomainValue.INTERVAL:
                if (min == undefined || max == undefined || min > max) {
                    // throw "invalid " + min + "/" + max + " min/max boundaries for 'interval' parameter definition domain"
                    let e: Message = new Message(MessageCode.ERROR_PARAMDOMAIN_INTERVAL_BOUNDS);
                    e.extraVar["minValue"] = min;
                    e.extraVar["maxValue"] = max;
                    throw e;
                }
                break;

            default:
                // en dehors du cas INTERVAL, on ne doit pas fournir de valeur
                if (min != undefined || max != undefined) {
                    let e: Message = new Message(MessageCode.ERROR_PARAMDOMAIN_INTERVAL_BOUNDS);
                    e.extraVar["minValue"] = min;
                    e.extraVar["maxValue"] = max;
                    throw e;
                }
                break;
        }
    }

    get domain() {
        return this._domain;
    }

    get minValue() {
        return this._minValue;
    }

    get maxValue() {
        return this._maxValue;
    }

    public static getDefaultBounds(d: ParamDomainValue): { min: number, max: number } {
        switch (d) {
            case ParamDomainValue.INTERVAL:
                let e: Message = new Message(MessageCode.ERROR_PARAMDOMAIN_INVALID);
                throw e;

            case ParamDomainValue.ANY:
            case ParamDomainValue.NOT_NULL:
                return { min: -Infinity, max: Infinity };

            case ParamDomainValue.POS:
                return { min: 1e-9, max: Infinity };

            case ParamDomainValue.POS_NULL:
                return { min: 0, max: Infinity };

            // default:
            //     throw "valeur de ParamDomainValue" + ParamDomainValue[d] + " non prise en charge";
        }
    }
}

/**
 * calculabilité du paramètre
 */
export enum ParamCalculability {
    /**
     * paramètre fixé (immuable, cad non modifiable après création)
     */
    NONE,

    /**
     * paramètre libre (modifiable)
     */
    FREE,

    /**
     * paramètre calculable analytiquement, par méthode de Newton, ...
     */
    EQUATION,

    /**
     * paramètre calculable par dichotomie
     */
    DICHO
}

/**
 * définition d'un paramètre
 */
export class ParamDefinition extends DefinedNumber {
    /**
     * symbole
     */
    private _symbol: string;

    /**
     * alias du symbole
     */
    //    public symbolAlias: string;

    /**
     * calculabilité
     */
    private _calc: ParamCalculability;

    /**
     * domaine de définition
     */
    private _domain: ParamDomain;

    /**
     * noeud de calcul parent
     */
    private _computeNodeType: ComputeNodeType

    /**
     * ancienne valeur du paramètre
     */
    // private _savedValue: number;

    /**
     * valeur du paramètre
     */
    //private _value: number;

    // private static _idGen: number = 0; // A VIRER
    // private _id: number; // A VIRER

    // constructor(s: string, d: ParamDomain, c: ParamCalculability, val: number = undefined) {
    constructor(nt: ComputeNodeType, s: string, d: ParamDomain | ParamDomainValue, val: number = undefined) {
        super(val);
        this._computeNodeType = nt;
        this._symbol = s;

        if (d instanceof ParamDomain)
            this._domain = d;
        else
            this._domain = new ParamDomain(<ParamDomainValue>d);

        this._calc = undefined;
        this.checkValue(val);
        // this._value = val;

        // this._id = ParamDefinition._idGen++; // A VIRER
        // console.log("constructor param " + this._symbol + " id=" + this._id); // A VIRER
    }

    // get id(): number {
    //     return this._id;
    // }

    /**
     * getter symbole
     */
    get symbol(): string {
        return this._symbol;
    }

    /**
     * méthodes de domaine
     */
    getDomain(): ParamDomain {
        return this._domain;
    }


    get computeNodeType(): ComputeNodeType {
        return this._computeNodeType;
    }

    /*
     * méthodes de calculabilité
     */

    /**
     * variable calculable par l'équation ?
     */
    isAnalytical(): boolean {
        return this.calculability == ParamCalculability.EQUATION;
    }

    get calculability(): ParamCalculability {
        if (this._calc == undefined) {
            //   throw "value of parameter '" + this._symbol + "' calculability is not defined";
            let e = new Message(MessageCode.ERROR_PARAMDEF_CALC_UNDEFINED);
            e.extraVar["symbol"] = this.symbol;
            throw e;
        }

        return this._calc;
    }

    set calculability(c: ParamCalculability) {
        this._calc = c;
    }

    /**
     * gestion de la valeur
     */
    get v(): number {
        // if (this._value == undefined)
        if (!this.isDefined) {
            //   throw new UndefinedError("value of '" + this._symbol + "' parameter is not defined");
            let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_UNDEFINED);
            e.extraVar["symbol"] = this.symbol;
            throw e;
        }
        // return this._value;
        return this.getValue();
    }

    set v(val: number) {
        //        if (val == undefined)
        //            console.log("warning : setting parameter '" + this._symbol + "' to undefined value (use undefine() instead)");

        if (this.calculability == ParamCalculability.NONE) {
            //   throw "value of '" + this._symbol + "' parameter cannot be changed";
            let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_FIXED);
            e.extraVar["symbol"] = this.symbol;
            throw e;
        }
        this.checkValue(val);
        // this._value = val;
        this.setValue(val);

        //        console.log("setting param " + this._symbol + " id=" + this._id + " to " + val); // A VIRER
    }

    // get uncheckedValue() {
    //     return this._value;
    // }

    checkValue(v: number) {
        let sDomain = ParamDomainValue[this._domain.domain];

        switch (this._domain.domain) {
            case ParamDomainValue.ANY:
                break;

            case ParamDomainValue.POS:
                if (v <= 0) {
                    //   throw "parameter '" + this._symbol + "' : invalid " + v + " value (cannot be <=0)";
                    let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_POS);
                    e.extraVar["symbol"] = this.symbol;
                    e.extraVar["value"] = v;
                    throw e;
                }
                break;

            case ParamDomainValue.POS_NULL:
                if (v < 0) {
                    //   throw "parameter '" + this._symbol + "' : invalid " + v + " value (cannot be <0)";
                    let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_POSNULL);
                    e.extraVar["symbol"] = this.symbol;
                    e.extraVar["value"] = v;
                    throw e;
                }
                break;

            case ParamDomainValue.NOT_NULL:
                if (v == 0) {
                    //   throw "parameter '" + this._symbol + "' : value cannot be 0";
                    let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_NULL);
                    e.extraVar["symbol"] = this.symbol;
                    throw e;
                }
                break;

            case ParamDomainValue.INTERVAL:
                let min = this._domain.minValue;
                let max = this._domain.maxValue;
                if (v < min || v > max) {
                    // throw "parameter '" + this._symbol + "' : value " + v + " is out of [" + min + ", " + max + "] interval";
                    let e = new Message(MessageCode.ERROR_PARAMDEF_VALUE_INTERVAL);
                    e.extraVar["symbol"] = this.symbol;
                    e.extraVar["value"] = v;
                    e.extraVar["minValue"] = min;
                    e.extraVar["maxValue"] = max;
                    throw e;
                }
                break;

            default:
                //throw "parameter '" + this._symbol + "' : non supported '" + sDomain + "' definition domain";
                let e = new Message(MessageCode.ERROR_PARAMDOMAIN_INVALID);
                e.extraVar["symbol"] = this.symbol;
                e.extraVar["domain"] = sDomain;
                throw e;
        }
    }

    // isDefined(): boolean {
    //     return this._value != undefined;
    // }


    // undefine() {
    //     this._value = undefined;
    //     // this._savedValue = undefined;
    // }

    /**
     * gestion du cache
     */
    /*
    saveValue() {
        if (this._value == undefined)
            throw "parameter '" + this._symbol + "' : cannot save 'undefined' value";
        if (this._savedValue != undefined)
            throw "parameter '" + this._symbol + "' : previous saved value " + this._savedValue + " has not been restored";
        this._savedValue = this._value;
    }

    restoreValue() {
        if (this._savedValue == undefined)
            throw "parameter '" + this._symbol + "' : value has not been saved";
        this._value = this._savedValue;
        this._savedValue = undefined;
    }
    */

    // toString(): string {
    //     if (this.isDefined())
    //         return "" + this._value;
    //     return "undefined";
    // }
}


/**
 * liste des paramètres d'une équation
 */
export abstract class ParamsEquation {
    protected _paramMap: { [key: string]: ParamDefinition } = {};

    protected addParamDefinition(p: ParamDefinition) {
        if (!this.hasParameter(p.symbol))
            this._paramMap[p.symbol] = p;
    }

    protected addParamDefinitions(ps: ParamsEquation) {
        for (let pi in ps._paramMap) {
            this.addParamDefinition(ps._paramMap[pi]);
        }
    }

    public get map(): { [key: string]: ParamDefinition } {
        return this._paramMap;
    }

    protected checkParametersCalculability() {
        let res = [];

        for (let ps in this._paramMap) {
            let p: ParamDefinition = this._paramMap[ps];
            if (p.calculability == undefined)
                res.push(p.symbol);
        }

        if (res.length > 0)
            throw 'Calculability of parameter(s) ' + res.toString() + ' has not been defined';
    }

    public hasParameter(name: string): boolean {
        for (let ps in this._paramMap) {
            let p: ParamDefinition = this._paramMap[ps];
            if (p.symbol == name)
                return true;
        }

        return false;
    }

    public getParameter(name: string): ParamDefinition {
        for (let ps in this._paramMap) {
            let p: ParamDefinition = this._paramMap[ps];
            if (p.symbol == name)
                return p;
        }

        throw 'ParamsEquation.getParameter() : invalid parameter name ' + name;
    }

    public getFirstAnalyticalParameter(): ParamDefinition {
        for (let ps in this._paramMap) {
            let p: ParamDefinition = this._paramMap[ps];
            if (p.isAnalytical())
                return p;
        }
        return undefined;
    }
}

/**
 * type de noeud de calcul
 */
export enum ComputeNodeType {
    CondDistri, LechaptCalmon,
    SectionParametree, // paramètres communs à toutes les sections paramétrées
    SectionTrapeze, SectionRectangle, SectionCercle, SectionPuissance,
    RegimeUniforme, // paramètres communs à toutes les sections paramétrées
    RegimeUniformeTrapeze, RegimeUniformeRectangle, RegimeUniformeCercle, RegimeUniformePuissance,
    CourbeRemous, // paramètres communs à toutes les courbes de remous
    CourbeRemousTrapeze, CourbeRemousRectangle, CourbeRemousCercle, CourbeRemousPuissance,
    Test
}

/**
 * noeud de calcul
 */
export abstract class ComputeNode extends Debug {
    protected _prms: ParamsEquation;

    protected abstract setParametersCalculability(): void;

    constructor(prms: ParamsEquation, dbg: boolean = false) {
        super(dbg);
        this._prms = prms;
        this.setParametersCalculability();
    }

    public getParameter(name: string): ParamDefinition {
        return this._prms.getParameter(name);
    }

    public getFirstAnalyticalParameter(): ParamDefinition {
        return this._prms.getFirstAnalyticalParameter();
    }
}
