import { ComputeNodeType, ParamDefinition, ParamDomain, ParamDomainValue, ParamCalculability } from "../param";
import { acSection, ParamsSection } from "./section_type";

/**
 * Paramètres de la section parabolique ou "puissance"
 */
export class ParamsSectionPuiss extends ParamsSection {
        private _k: ParamDefinition; // Coefficient de forme compris entre 0 et 1

        constructor(rk: number, rY: number, rLargeurBerge: number, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number,
                nodeType: ComputeNodeType = ComputeNodeType.SectionPuissance,
                defaultNodeType: ComputeNodeType = ComputeNodeType.SectionParametree) {
                super(rY, rLargeurBerge, rKs, rQ, rIf, rPrec, rYB, defaultNodeType);
                this._k = new ParamDefinition(nodeType, 'k', new ParamDomain(ParamDomainValue.INTERVAL, 0, 1), rk);

                this.addParamDefinition(this._k);
        }

        /**
        * Coefficient de forme compris entre 0 et 1
        */
        get k(): ParamDefinition {
                return this._k;
        }
}

/**
 * Calculs de la section parabolique ou "puissance"
 */
export class cSnPuiss extends acSection {
        protected nbDessinPoints = 50;

        constructor(prms: ParamsSectionPuiss, dbg: boolean = false) {
                super(prms, dbg);
        }

        protected setParametersCalculability() {
                super.setParametersCalculability();
                this.prms.k.calculability = ParamCalculability.DICHO;
        }

        get prms(): ParamsSectionPuiss {
                return <ParamsSectionPuiss>this._prms;
        }

        /**
         * Calcul de Lambda (mais on garde la routine Alpha commune avec la section circulaire)
         * @return Lambda
         */
        protected Calc_Alpha(): number {
                return this.prms.LargeurBerge.v / Math.pow(this.prms.YB.v, this.prms.k.v);
        }

        /**
         * Calcul de la largeur au miroir.
         * @return B
         */
        protected Calc_B(): number {
                if (this.prms.Y.v >= this.prms.YB.v)
                        return this.prms.LargeurBerge.v;

                return this.Calc("Alpha") * Math.pow(this.prms.Y.v, this.prms.k.v);
        }

        /**
         * Calcul du périmètre mouillé.
         * @return B
         */
        protected Calc_P(): number {
                var n = 100; /// Le nombre de partie pour le calcul de l'intégrale
                var Lambda2 = Math.pow(this.Calc("Alpha"), 2);
                var P = 0; /// Le périmètre à calculer
                var Previous = 0;
                for (var i = 1; i <= n; i++) {
                        var Current = Math.pow(this.prms.Y.v * i / n, this.prms.k.v) / 2;
                        P += Math.sqrt(Math.pow(n, -2) + Lambda2 * Math.pow(Current - Previous, 2));
                        Previous = Current;
                }
                P *= 2;
                return P;
        }

        /**
         * Calcul de la surface mouillée.
         * @return S
         */
        protected Calc_S(): number {
                let k = this.prms.k.v;
                return this.Calc("Alpha") * Math.pow(this.prms.Y.v, k + 1) / (k + 1);
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected Calc_dP(): number {
                let k = this.prms.k.v;
                return 2 * Math.sqrt(1 + Math.pow(k * this.Calc("Alpha") / 2, 2) * Math.pow(this.prms.Y.v, 2 * (k - 1)));
        }

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
         * @return dB
         */
        protected Calc_dB(): number {
                let k = this.prms.k.v;
                return this.Calc("Alpha") * k * Math.pow(this.prms.Y.v, k - 1);
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): number {
                let k = this.prms.k.v;
                return this.Calc("Alpha") * Math.pow(this.prms.Y.v, k + 2) / ((k + 1) * (k + 2));
        }

        /**
         * Calcul de la dérivée distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): number {
                let k = this.prms.k.v;
                let Y = this.prms.Y.v;
                let SYg = this.Calc("dAlpha") * Math.pow(Y, k + 2) + this.Calc("Alpha") * Math.pow(Y, k + 1) * (k + 2);
                return SYg / ((k + 1) * (k + 2));
        }
}