import { ComputeNodeType, ParamDefinition, ParamDomainValue, ParamCalculability } from "../param";
import { acSection, ParamsSection } from "./section_type";
import { cLog } from "../util/log";


export class ParamsSectionTrapez extends ParamsSection {
        private _LargeurFond: ParamDefinition; // Largeur au fond
        private _Fruit: ParamDefinition; // Fruit des berges

        constructor(rLargeurFond: number, rFruit: number, rY: number, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number,
                nodeType: ComputeNodeType = ComputeNodeType.SectionTrapeze,
                defaultNodeType: ComputeNodeType = ComputeNodeType.SectionParametree) {
                super(rY, undefined, rKs, rQ, rIf, rPrec, rYB, defaultNodeType);
                this._LargeurFond = new ParamDefinition(nodeType, 'LargeurFond', ParamDomainValue.POS_NULL, rLargeurFond);
                this._Fruit = new ParamDefinition(nodeType, 'Fruit', ParamDomainValue.POS_NULL, rFruit);

                this.addParamDefinition(this._LargeurFond);
                this.addParamDefinition(this._Fruit);
        }

        /**
        * Largeur au fond
        */
        get LargeurFond(): ParamDefinition {
                return this._LargeurFond;
        }

        /**
        * Fruit des berges
        */
        get Fruit(): ParamDefinition {
                return this._Fruit;
        }
}


/**
 * Calculs de la section trapézoïdale
 */
export class cSnTrapez extends acSection {
        constructor(prms: ParamsSectionTrapez, dbg: boolean = false) {
                super(prms, dbg);
        }

        protected setParametersCalculability() {
                super.setParametersCalculability();
                this.prms.LargeurFond.calculability = ParamCalculability.DICHO;
                this.prms.Fruit.calculability = ParamCalculability.DICHO;
        }

        get prms(): ParamsSectionTrapez {
                return <ParamsSectionTrapez>this._prms;
        }

        private isDebordement(): boolean {
                return this.prms.Y.v > this.prms.YB.v;
        }

        // Calc_B(bBerge = false): number {
        protected Calc_B(): number {
                // if (!bBerge && this.isDebordement())
                if (this.isDebordement())
                        //return this.prms.LargeurBerge.v;
                        return this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.YB.v;

                return this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
        }

        /**
         * Calcul du périmètre mouillé
         * @return Périmètre mouillé (m)
         */
        protected Calc_P(): number {
                if (this.isDebordement())
                        return this.CalcGeo("P") + super.Calc_P_Debordement(this.prms.Y.v - this.prms.YB.v);

                return this.prms.LargeurFond.v + 2 * Math.sqrt(1 + Math.pow(this.prms.Fruit.v, 2)) * this.prms.Y.v;
        }

        /**
         * Calcul de la surface mouillée
         * @return Surface mouillée (m2)
         */
        protected Calc_S(): number {
                if (this.isDebordement())
                        return this.CalcGeo("S") + super.Calc_S_Debordement(this.prms.Y.v - this.prms.YB.v);

                return this.prms.Y.v * (this.prms.LargeurFond.v + this.prms.Fruit.v * this.prms.Y.v);
        }

        /**
         * Calcul de dérivée de la surface hydraulique par rapport au tirant d'eau.
         * @return dS
         */
        protected Calc_dS(): number {
                if (this.isDebordement())
                        return super.Calc_dS();

                return this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected Calc_dP(): number {
                if (this.isDebordement())
                        return super.Calc_dP_Debordement();

                return 2 * Math.sqrt(1 + this.prms.Fruit.v * this.prms.Fruit.v);
        }

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
         * @return dB
         */
        protected Calc_dB(): number {
                if (this.isDebordement())
                        return super.Calc_dB_Debordement();

                return 2 * this.prms.LargeurFond.v * this.prms.Fruit.v;
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): number {
                return (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3) * Math.pow(this.prms.Y.v, 2);
        }

        /**
         * Calcul de la dérivée de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): number {
                var SYg = this.prms.Fruit.v / 3 * Math.pow(this.prms.Y.v, 2);
                SYg += (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3) * 2 * this.prms.Y.v;
                return SYg;
        }
}