import { XOR } from "../base";
import { cLog } from "../util/log";
import { MessageCode, Message } from "../util/message";
import { ComputeNodeType, ComputeNode, ParamDefinition, ParamDomainValue, ParamCalculability, ParamsEquation } from "../param";
import { cHautCritique, cHautNormale, cHautCorrespondante, cHautConjuguee } from "./hauteur";

/**
 * Gestion des Paramètres du canal (hors section)
 */
export abstract class cParamsCanal extends ParamsEquation {
        public static readonly G: number = 9.81;/// Constante de gravité

        private _Ks: ParamDefinition; // Strickler
        private _Q: ParamDefinition; // Débit
        private _If: ParamDefinition;  // Pente du fond
        private _Prec: ParamDefinition;  // Précision de calcul et d'affichage
        private _iPrec: ParamDefinition;  // Précision en nombre de décimales
        private _YB: ParamDefinition;  // Hauteur de berge

        constructor(nodeType: ComputeNodeType, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number) {
                super();
                this._Ks = new ParamDefinition(nodeType, 'Ks', ParamDomainValue.POS, rKs);
                this._Q = new ParamDefinition(nodeType, 'Q', ParamDomainValue.POS_NULL, rQ);
                this._If = new ParamDefinition(nodeType, 'If', ParamDomainValue.ANY, rIf);
                this._Prec = new ParamDefinition(nodeType, 'Prec', ParamDomainValue.POS, rPrec);
                this._iPrec = new ParamDefinition(nodeType, 'iPrec', ParamDomainValue.ANY, Math.round(-Math.log(rPrec) / Math.log(10)));
                this._YB = new ParamDefinition(nodeType, 'YB', ParamDomainValue.POS, rYB);

                this.addParamDefinition(this._Ks);
                this.addParamDefinition(this._Q);
                this.addParamDefinition(this._If);
                this.addParamDefinition(this._Prec);
                this.addParamDefinition(this._iPrec);
                this.addParamDefinition(this._YB);
        }

        /**
         * Strickler
         */
        get Ks(): ParamDefinition {
                return this._Ks;
        }

        /**
         * Débit
         */
        get Q(): ParamDefinition {
                return this._Q;
        }

        /**
         * Pente du fond
         */
        get If(): ParamDefinition {
                return this._If;
        }

        /**
         * Précision de calcul et d'affichage
         */
        get Prec(): ParamDefinition {
                return this._Prec;
        }

        /**
         * Précision en nombre de décimales
         */
        get iPrec(): ParamDefinition {
                return this._iPrec;
        }

        /**
         * Hauteur de berge
         */
        get YB(): ParamDefinition {
                return this._YB;
        }

        // /**
        //  * Longueur du bief
        //  */
        // get Long() {
        //         return this._Long;
        // }
}


export abstract class ParamsSection extends cParamsCanal {
        private _Y: ParamDefinition;          // Tirant d'eau
        private _LargeurBerge: ParamDefinition; // largeur au débordement

        constructor(rY: number, rLargeurBerge: number, rKs: number, rQ: number, rIf: number, rPrec: number, rYB: number,
                nodeType: ComputeNodeType = ComputeNodeType.SectionParametree) {
                super(nodeType, rKs, rQ, rIf, rPrec, rYB);
                this._Y = new ParamDefinition(nodeType, 'Y', ParamDomainValue.POS_NULL, rY);
                this._LargeurBerge = new ParamDefinition(nodeType, 'LargeurBerge', ParamDomainValue.POS_NULL, rLargeurBerge);

                this.addParamDefinition(this._Y);
                this.addParamDefinition(this._LargeurBerge);
        }

        /**
        * Tirant d'eau
        */
        get Y(): ParamDefinition {
                return this._Y;
        }

        /**
        * Largeur au débordement
        */
        get LargeurBerge(): ParamDefinition {
                return this._LargeurBerge;
        }
}

/**
 * Gestion commune pour les différents types de section.
 * Comprend les formules pour la section rectangulaire pour gérer les débordements
 */
export abstract class acSection extends ComputeNode {
        [key: string]: any; // pour pouvoir faire this['methode]();

        public HautCritique: number;  /// Tirant d'eau critique
        public HautNormale: number;   /// Tirant d'eau normal
        protected oLog: cLog; /// Pour l'affichage du journal de calcul

        /**
         * true si la section est fermée (on considère alors qu'il existe une fente de Preissmann)
         */
        protected bSnFermee: boolean = false;

        /**
         * Tableau contenant les données dépendantes du tirant d'eau this->rY.
         *
         * Les clés du tableau peuvent être :
         * - S : la surface hydraulique
         * - P : le périmètre hydraulique
         * - R : le rayon hydraulique
         * - B : la largeur au miroir
         * - J : la perte de charge
         * - Fr : le nombre de Froude
         * - dP : la dérivée de P par rapport Y
         * - dR : la dérivée de R par rapport Y
         * - dB : la dérivée de B par rapport Y
         */
        private arCalc: { [key: string]: number } = {};
        protected arCalcGeo: { [key: string]: number } = {}; /// Données ne dépendant pas de la cote de l'eau
        private Y_old: number; /// Mémorisation du tirant d'eau pour calcul intermédiaire
        private Calc_old = {}; /// Mémorisation des données hydrauliques pour calcul intermédiaire

        /**
         * itérations max pour Newton
         */
        private _newtonMaxIter: number = 50;

        /**
         * Construction de la classe.
         * Calcul des hauteurs normale et critique
         */
        constructor(prms: ParamsSection, dbg: boolean = false) {
                super(prms, dbg);
                this.oLog = new cLog();

                // this._v = oP.v;  // copie la référence au tableau
                // this.prms.Y.v = 0;
                // this.prms.LargeurBerge.v = 0;

                // this.CalcGeo("B");  // à ce point, il manque des paramètres pour certaines sections (ex : fruit pour trapèze)
        }

        protected setParametersCalculability() {
                this.prms.Ks.calculability = ParamCalculability.DICHO;
                this.prms.Q.calculability = ParamCalculability.DICHO;
                this.prms.If.calculability = ParamCalculability.DICHO;
                this.prms.Prec.calculability = ParamCalculability.NONE;
                this.prms.iPrec.calculability = ParamCalculability.NONE;
                this.prms.YB.calculability = ParamCalculability.DICHO;
                this.prms.Y.calculability = ParamCalculability.DICHO;
                this.prms.LargeurBerge.calculability = ParamCalculability.DICHO;
        }

        get prms(): ParamsSection {
                return <ParamsSection>this._prms;
        }

        public clone(): acSection {
                return Object.create(this);
        }

        public set newtonMaxIter(n: number) {
                this._newtonMaxIter = n;
        }

        public get log(): cLog {
                return this.oLog;
        }

        /**
         * Efface toutes les données calculées pour forcer le recalcul
         * @param bGeo Réinitialise les données de géométrie aussi
         */
        public Reset(bGeo = true) {
                this.debug("reset(" + bGeo + ")");
                this.arCalc = {};
                if (bGeo) {
                        // this.debug("reset Geo");
                        this.arCalcGeo = {};
                        this.bSnFermee = false;
                }
                // this.debug("reset arCalc :");
                // this.debug(this.arCalc);
                // this.debug("reset arCalcGeo :");
                // this.debug(this.arCalcGeo);
        }

        /**
         * Mémorise les données hydraulique en cours ou les restitue
         * @param bMem true pour mémorisation, false pour restitution
         */
        private Swap(bMem: boolean) {
                if (bMem) {
                        this.debug("save Y=" + this.prms.Y.toString());
                        this.Y_old = this.prms.Y.uncheckedValue;
                        this.Calc_old = this.arCalc;
                }
                else {
                        this.debug("restore Y=" + this.Y_old);
                        this.prms.Y.v = this.Y_old;
                        this.arCalc = this.Calc_old;
                        this.Calc_old = {};
                }
        }

        /**
         * Calcul des données à la section
         * @param sDonnee Clé de la donnée à calculer (voir this->arCalc)
         * @param rY valeur de Y à utiliser
         * @return la donnée calculée
         */
        public Calc(sDonnee: string, rY: number = undefined): number {
                this.debug("in Calc(" + sDonnee + ", rY=" + rY + ") old " + sDonnee + "=" + this.arCalc[sDonnee]);
                this.debug("this.Y=" + this.prms.Y.toString());

                if (rY != undefined && (!this.prms.Y.isDefined || rY != this.prms.Y.v)) {
                        this.prms.Y.v = rY;
                        // On efface toutes les données dépendantes de Y pour forcer le calcul
                        this.Reset(false);
                }
                if (this.arCalc[sDonnee] == undefined) {
                        // La donnée a besoin d'être calculée
                        switch (sDonnee) {
                                case 'I-J': // Variation linéaire de l'énergie spécifique (I-J) en m/m
                                        this.arCalc[sDonnee] = this.prms.If.v - this.Calc("J");
                                        break;

                                default:
                                        let methode = 'Calc_' + sDonnee;

                                        /* !!!!!
                                           si on réunit les 2 lignes suivantes ( this.arCalc[sDonnee] = this[methode](); ),
                                           il arrive que this.arCalc[sDonnee] soit affecté à undefined alors que this[methode]() renvoie
                                           une valeur bien définie
                                           !!!!!
                                         */
                                        let res: number = this[methode]();
                                        this.arCalc[sDonnee] = res;
                                        break;
                        }
                        this.debug("Calc(" + sDonnee + ") resultat -> " + this.arCalc[sDonnee]);
                }
                else
                        this.debug("Calc(" + sDonnee + ") cache= " + this.arCalc[sDonnee]);

                this.debug('this.Y=' + this.prms.Y.toString());

                return this.arCalc[sDonnee];
        }

        /**
         * Calcul des données uniquement dépendantes de la géométrie de la section
         * @param sDonnee Clé de la donnée à calculer (voir this->arCalcGeo)
         * @return la donnée calculée
         */
        protected CalcGeo(sDonnee: string): number {
                this.debug("in CalcGeo(" + sDonnee + ") old " + sDonnee + "=" + this.arCalcGeo[sDonnee]);
                this.debug("this.Y=" + this.prms.Y.toString());

                // Si la largeur aux berges n'a pas encore été calculée, on commence par ça
                if (sDonnee != "B" && this.arCalcGeo["B"] == undefined)
                        this.CalcGeo("B");

                if (this.arCalcGeo[sDonnee] == undefined) {
                        // La donnée a besoin d'être calculée
                        this.Swap(true); // On mémorise les données hydrauliques en cours
                        this.Reset(false);
                        this.prms.Y.v = this.prms.YB.v;
                        switch (sDonnee) {
                                case "B": // Largeur aux berges
                                        this.arCalcGeo[sDonnee] = this.Calc_B();
                                        if (this.arCalcGeo[sDonnee] < this.prms.YB.v / 100) {
                                                // Section fermée
                                                this.bSnFermee = true;
                                                // On propose une fente de Preissmann égale à 1/100 de la hauteur des berges
                                                this.arCalcGeo[sDonnee] = this.prms.YB.v / 100;
                                        }
                                        this.prms.LargeurBerge.v = this.arCalcGeo[sDonnee];
                                        break;

                                default:
                                        let methode = "Calc_" + sDonnee;
                                        // this.arCalcGeo[sDonnee] = this[methode]();
                                        let res = this[methode]();
                                        this.arCalcGeo[sDonnee] = res;
                                        break;
                        }
                        this.Swap(false); // On restitue les données hydrauliques en cours
                        this.debug("CalcGeo(" + sDonnee + ") resultat -> " + this.arCalcGeo[sDonnee]);
                }
                else
                        this.debug("CalcGeo(" + sDonnee + ") cache= " + this.arCalcGeo[sDonnee]);

                this.debug('this.Y=' + this.prms.Y.toString());

                return this.arCalcGeo[sDonnee];
        }

        /**
         * Calcul de la surface hydraulique.
         * @return La surface hydraulique
         */
        protected abstract Calc_S(): number;


        protected Calc_dS(): number {
                return this.Calc("B"); // largeur au miroir
        }

        /**
         *  calcul de la surface hydraulique en cas de débordement
         * @param Y hauteur d'eau au dela de la berge
         **/
        protected Calc_S_Debordement(Y: number): number {
                // this.debug('section->CalcS(rY=' + Y + ') LargeurBerge=' + this.prms.LargeurBerge);
                this.debug('section->CalcS(rY=' + Y + ') LargeurBerge=' + this.CalcGeo("B"));
                // return Y * this.prms.LargeurBerge.v;
                return Y * this.CalcGeo("B");
        }

        /**
         * Calcul de la dérivée surface hydraulique en cas de débordement
         * @return La dérivée de la surface hydraulique en cas de débordement
         */
        private Calc_dS_Debordement(): number {
                // return this.prms.LargeurBerge.v;
                return this.CalcGeo("B");
        }

        /**
         * Calcul du périmètre hydraulique.
         * @return Le périmètre hydraulique
         */
        protected abstract Calc_P(): number;

        /**
         * Calcul du périmètre hydraulique en cas de débordement
         * @param Y hauteur d'eau au dela de la berge
         */
        protected Calc_P_Debordement(Y: number): number {
                return 2 * Y;
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected abstract Calc_dP(): number;

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau en cas de débordement
         * @return la dérivée du périmètre hydraulique par rapport au tirant d'eau en cas de débordement
         */
        protected Calc_dP_Debordement(): number {
                return 2;
        }

        /**
         * Calcul du rayon hydraulique.
         * @return Le rayon hydraulique
         */
        private Calc_R(): number {
                let P = this.Calc("P");
                if (P != 0)
                        return this.Calc("S") / P;

                return Infinity;
        }

        /**
         * Calcul de dérivée du rayon hydraulique par rapport au tirant d'eau.
         * @return dR
         */
        private Calc_dR(): number {
                let P = this.Calc("P");
                if (P != 0)
                        return ((this.Calc("B") * P - this.Calc("S") * this.Calc("dP")) / Math.pow(P, 2));

                return 0;
        }

        /**
         * Calcul de la largeur au miroir.
         * @return La largeur au miroir
         */
        protected abstract Calc_B(): number;


        /**
           * Calcul de la largeur au miroir en cas de débordement
           * @return La largeur au miroir en cas de débordement
           */
        protected Calc_B_Debordement(): number {
                return this.prms.LargeurBerge.v;
                // return this.CalcGeo("B");
        }

        /**
          * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
          * @return dB
          */
        protected abstract Calc_dB(): number;

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau en cas de débordement
         * @return la dérivée de la largeur au miroir par rapport au tirant d'eau en cas de débordement
         */
        protected Calc_dB_Debordement(): number {
                return 0;
        }

        /**
         * Calcul de la perte de charge par la formule de Manning-Strickler.
         * @return La perte de charge
         */
        private Calc_J(): number {
                let R = this.Calc("R");
                if (R != 0)
                        return Math.pow(this.Calc("V") / this.prms.Ks.v, 2) / Math.pow(R, 4 / 3);

                return Infinity;
        }

        /**
         * Calcul du nombre de Froude.
         * @return Le nombre de Froude
         */
        private Calc_Fr(): number {
                let S = this.Calc("S");
                if (S != 0)
                        return this.prms.Q.v / S * Math.sqrt(this.Calc("B") / S / cParamsCanal.G);

                return Infinity;
        }

        /**
         * Calcul de dy/dx
         */
        private Calc_dYdX(Y: number): number {
                // L'appel à Calc("J') avec Y en paramètre réinitialise toutes les données dépendantes de la ligne d'eau
                return - (this.prms.If.v - this.Calc("J", Y) / (1 - Math.pow(this.Calc("Fr", Y), 2)));
        }

        /**
         * Calcul de la vitesse moyenne.
         * @return Vitesse moyenne
         */
        private Calc_V(): number {
                let S = this.Calc("S");
                if (S != 0)
                        return this.prms.Q.v / S;

                return Infinity;
        }

        /**
         * Calcul de la charge spécifique.
         * @return Charge spécifique
         */
        private Calc_Hs(): number {
                return this.prms.Y.v + Math.pow(this.Calc("V"), 2) / (2 * cParamsCanal.G);
        }

        /**
         * Calcul de la charge spécifique critique.
         * @return Charge spécifique critique
         */
        private Calc_Hsc(): number {
                this.Swap(true); // On mémorise les données hydrauliques en cours
                // On calcule la charge avec la hauteur critique
                var Hsc = this.Calc("Hs", this.CalcGeo("Yc"));
                // On restitue les données initiales
                this.Swap(false);
                return Hsc;
        }

        /**
         * Calcul du tirant d'eau critique.
         * @return tirant d'eau critique
         */
        private Calc_Yc(): number {
                let hautCritique = new cHautCritique(this, this._newtonMaxIter, this.DBG);
                this.HautCritique = hautCritique.Newton(this.prms.YB.v);
                if (this.HautCritique == undefined || !hautCritique.hasConverged()) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCRITIQUE);
                        this.oLog.add(m);
                }
                return this.HautCritique;
        }

        /**
         * Calcul du tirant d'eau normal.
         * @return tirant d'eau normal
         */
        private Calc_Yn(): number {
                this.debug("in calc_Yn");
                if (this.prms.If.v <= 0) {
                        let m = new Message(MessageCode.ERROR_SECTION_PENTE_NEG_NULLE_HNORMALE_INF);
                        this.oLog.add(m);
                        return undefined;
                }

                let oHautNormale = new cHautNormale(this, this._newtonMaxIter, this.DBG);
                let res = oHautNormale.Newton(this.CalcGeo("Yc"));
                if (res == undefined || !oHautNormale.hasConverged()) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HNORMALE);
                        this.oLog.add(m);
                }

                return res;
        }

        /**
         * Calcul du tirant d'eau fluvial.
         * @return tirant d'eau fluvial
         */
        private Calc_Yf(): number {
                if (this.prms.Y.v > this.CalcGeo("Yc"))
                        return this.prms.Y.v;

                var oHautCorrespondante = new cHautCorrespondante(this, this._newtonMaxIter, this.DBG);
                let res = oHautCorrespondante.Newton(this.Calc("Yc") * 2);
                if (res == undefined || !oHautCorrespondante.hasConverged()) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HFLU);
                        this.oLog.add(m);
                }

                return res;
        }

        /**
         * Calcul du tirant d'eau torrentiel.
         * @return tirant d'eau torrentiel
         */
        private Calc_Yt(): number {
                if (this.prms.Y.v < this.CalcGeo("Yc"))
                        return this.prms.Y.v;

                var oHautCorrespondante = new cHautCorrespondante(this, this._newtonMaxIter, this.DBG);
                let res = oHautCorrespondante.Newton(this.CalcGeo("Yc") / 2);
                if (res == undefined || !oHautCorrespondante.hasConverged()) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HTOR);
                        this.oLog.add(m);
                }

                return res;
        }

        /**
         * Calcul du tirant d'eau conjugué.
         * @return tirant d'eau conjugué
         */
        private Calc_Yco(): number {
                this.Swap(true);
                var oHautConj = new cHautConjuguee(this, this._newtonMaxIter, this.DBG);
                // Choisir une valeur initiale du bon côté de la courbe
                if (this.Calc("Fr") < 1) {
                        // Ecoulement fluvial, on cherche la conjuguée à partir du tirant d'eau torrentiel
                        var Y0 = this.Calc("Yt");
                }
                else {
                        // Ecoulement torrentiel, on cherche la conjuguée à partir du tirant d'eau fluvial
                        Y0 = this.Calc("Yf");
                }
                let Yco = oHautConj.Newton(Y0);
                if (Yco == undefined || !oHautConj.hasConverged()) {
                        let m = new Message(MessageCode.ERROR_SECTION_NON_CONVERGENCE_NEWTON_HCONJUG);
                        this.oLog.add(m);
                }

                this.Swap(false);
                return Yco;
        }

        /**
         * Calcul de la contrainte de cisaillement.
         * @return contrainte de cisaillement
         */
        private Calc_Tau0(): number {
                return 1000 * cParamsCanal.G * this.Calc("R") * this.Calc("J");
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): number {
                // return Math.pow(this.prms.Y.v, 2) * this.prms.LargeurBerge.v / 2;
                return Math.pow(this.prms.Y.v, 2) * this.CalcGeo("B") / 2;
        }

        /**
         * Calcul de la dérivée distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): number {
                // return this.prms.Y.v * this.prms.LargeurBerge.v;
                return this.prms.Y.v * this.CalcGeo("B");
        }

        /**
         * Calcul de l'impulsion hydraulique.
         * @return Impulsion hydraulique
         */
        private Calc_Imp(): number {
                return 1000 * (this.prms.Q.v * this.Calc("V") + cParamsCanal.G * this.Calc("SYg"));
        }

        /**
         * Calcul de l'angle Alpha entre la surface libre et le fond pour les sections circulaires.
         * @return Angle Alpha pour une section circulaire, 0 sinon.
         */
        protected Calc_Alpha(): number {
                return 0;
        }

        /**
         * Calcul de la dérivée de l'angle Alpha entre la surface libre et le fond pour les sections circulaires.
         * @return Dérivée de l'angle Alpha pour une section circulaire, 0 sinon.
         */
        protected Calc_dAlpha(): number {
                return 0;
        }
}
